#!/usr/bin/env
"""
Protein-specific methods for .
"""

import itertools
import logging

import pandas as pd

from ..uniprot.id_mapping import id_mapper_context


LOGGER = logging.getLogger(__name__)


def get_uniprot_id(row, bqm):
    """
    Get the UniProt ID for a row if it exists.

    Args:
        row:
            The row from the dataframe of entities.

        bqm:
            An instance of :py:class:`~.query_manager.BiopaxQueryManager`.

    Returns:
        The UniProt ID, or None.
    """
    xrefs = row["xrefs"]
    if not xrefs or pd.isna(xrefs):
        return None
    for xref in xrefs.split(bqm.ITEM_DELIMITER):
        name, value = xref.split(bqm.FIELD_DELIMITER)
        if "UniProt" in name:
            return value
    return None


def _map_members_to_uniprot_ids(row, bqm, uniprot_dict):
    """
    Application function for map_members_to_uniprot_ids.
    """
    members = row["members"]
    if not members or pd.isna(members):
        return None
    uniprot_ids = set()
    for member in members.split(bqm.ITEM_DELIMITER):
        uniprot_id = uniprot_dict.get(member)
        if uniprot_id and not pd.isna(uniprot_id):
            uniprot_ids.add(uniprot_id)
    if uniprot_ids:
        return bqm.ITEM_DELIMITER.join(sorted(uniprot_ids))
    return None


def map_members_to_uniprot_ids(prot_ents, bqm):
    """
    Get a series with UniProt IDs of the members in the members column.

    Args:
        prot_ents:
            The protein entity dataframe.

        bqm:
            An instance of :py:class:`~.query_manager.BiopaxQueryManager`.

    Returns:
        A Pandas series with the UniProt IDs.
    """
    uniprot_dict = prot_ents.set_index("entity").to_dict()["uniprot"]
    return prot_ents.apply(
        _map_members_to_uniprot_ids, axis=1, args=(bqm, uniprot_dict)
    )


def _map_uniprot_ids_to_gene_id(value, bqm, id_map):
    """
    Application function for map_uniprot_ids_to_gene_ids.

    Args:
        value:
            The column value, containing 0 or more UniProt IDs separated by the
            bqm's item delimiter.

        bqm:
            An instance of :py:class:`~.query_manager.BiopaxQueryManager`.

        id_map:
            A dict mapping UniProt IDs to some gene ID.

    Returns:
        The mapped value.
    """
    if not value:
        return ""
    delimiter = bqm.ITEM_DELIMITER
    return delimiter.join(id_map.get(v, "") for v in value.split(delimiter))


def map_uniprot_ids_to_gene_ids(prot_ents, bqm):
    """
    Map UniProt IDs to different gene IDs (GeneID, Ensembl).

    Args:
        prot_ents:
            The protein entity dataframe.

        bqm:
            An instance of :py:class:`~.query_manager.BiopaxQueryManager`.

    Returns:
        The input dataframe with additional columns for the gene IDs. They will
        follow the format of the UniProt ID columns. For example, The "uniprot"
        column will map to a "geneid" column, and the "member_uniprot" will map
        to a "member_geneid" column. Multiple UniProt IDs separated by the
        BiopaxQueryManager's item delimiter will map to multiple gene IDs
        separated by the same delimiter. Missing values will map to empty
        strings.
    """
    non_null_uniprot = ~prot_ents["uniprot"].isnull()
    non_null_member_uniprot = ~prot_ents["member_uniprot"].isnull()

    # Collect all of the UniProt IDs.
    uniprot_ids = set(prot_ents[non_null_uniprot]["uniprot"].unique())
    member_uniprot_ids = (
        prot_ents["member_uniprot"][non_null_member_uniprot]
        .apply(lambda x: x.split(bqm.ITEM_DELIMITER))
        .tolist()
    )
    uniprot_ids.update(itertools.chain.from_iterable(member_uniprot_ids))

    with id_mapper_context(bqm.cache_man) as id_mapper:
        for gene_id in ("GeneID", "Ensembl"):
            id_map = id_mapper.map_ids("UniProtKB_AC-ID", gene_id, uniprot_ids)
            for prefix in ("", "member_"):
                prot_ents[f"{prefix}{gene_id}".lower()] = prot_ents[
                    f"{prefix}uniprot"
                ].apply(_map_uniprot_ids_to_gene_id, args=(bqm, id_map))


def get_protein_entities(bqm):
    """
    Get the dataframe of protein entities.

    Args:
        bqm:
            An instance of :py:class:`~.query_manager.BiopaxQueryManager`.

    Returns:
        A Pandas DataFrame.
    """
    phys_ents = bqm.query_physical_entities().merge(
        bqm.query_locations(), how="left", on="entity"
    )
    prot_ents = phys_ents[phys_ents["entity"].str.contains("Protein")].copy()
    if prot_ents.shape[0] != prot_ents["entity"].nunique():
        LOGGER.warning("Duplicate protein entities found in dataframe")

    no_display_name = prot_ents["display_names"].isna()
    if no_display_name.any():
        n_missing = no_display_name.sum()
        LOGGER.warning(
            "%d entit%s missing a display name",
            n_missing,
            "y is" if n_missing == 1 else "ies are",
        )
        no_other_name = prot_ents[no_display_name]["names"].isna()
        if no_other_name.any():
            n_missing = no_other_name.sum()
            LOGGER.warning(
                "%d entit%s have no name at all",
                n_missing,
                "y" if n_missing == 1 else "ies",
            )

    multiple_display_name = prot_ents["display_names"].str.contains(bqm.ITEM_DELIMITER)
    if multiple_display_name.any():
        n_multi = multiple_display_name.sum()
        LOGGER.warning(
            "%d entit%s have multiple display names",
            n_multi,
            "y" if n_multi == 1 else "ies",
        )

    prot_ents["uniprot"] = prot_ents.apply(get_uniprot_id, axis=1, args=(bqm,))
    prot_ents["member_uniprot"] = map_members_to_uniprot_ids(prot_ents, bqm)

    map_uniprot_ids_to_gene_ids(prot_ents, bqm)

    return prot_ents



#!/usr/bin/env python3
"""
Cached data management.
"""

import logging
import pathlib
import shutil

import xdg.BaseDirectory

from .common import NAME
from .config import Config
from .exception import MetabographConfigError

LOGGER = logging.getLogger(__name__)


class CacheManager:
    """
    Manage cached data.
    """

    def __init__(self, config: Config):
        self.config = config

    @property
    def cache_dir(self) -> pathlib.Path:
        """
        The cache directory path.

        Returns:
            The path to this applications cache directory.
        """
        try:
            path = self.config.cache.path
            if not path:
                raise MetabographConfigError("Undefined cache path")
        except AttributeError:
            return pathlib.Path(xdg.BaseDirectory.xdg_cache_home).resolve() / NAME
        return self.config.resolve_path(path)

    def get_path(
        self, subpath: str | pathlib.Path, directory: bool = False
    ) -> pathlib.Path:
        """
        Get a path to a subpath in the cache directory. The parent directory
        will be created if missing.

        Args:
            subpath:
                A subpath to interpret relative to the cache directory.

            directory:
                If True, the subpath will be created as a directory.

        Returns:
            A path within the cache directory.
        """
        cache_dir = self.cache_dir
        path = cache_dir / subpath
        # Check that the subpath lies within the cache directory.
        if not path.is_relative_to(cache_dir):
            raise MetabographConfigError(
                f'The subpath "{subpath}" lies outside of the cache directory.'
            )
        # Create the required directories.
        (path if directory else path.parent).mkdir(parents=True, exist_ok=True)
        return path

    def clear(self):
        """
        Clear all cached data.
        """
        cache_dir = self.cache_dir
        LOGGER.info("Removing cache directory: %s", cache_dir)
        try:
            shutil.rmtree(cache_dir)
        except FileNotFoundError:
            pass



#!/usr/bin/env python3
"""
Graph configuration.
"""

import dataclasses
import logging
import pathlib
from typing import Any, Optional

from dataclass_documenter import DataclassDocumenter

import yaml


LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class CacheData:
    """
    Cache data configuration.

    Parameters:
        path:
            The path to a cache directory. If unset, the standard XDG user cache
            directory will be used.

        timeout:
            The timeout for the cached data. Data will be cleared from the cache
            after this timeout. If unset, cached data will not automatically
            time out.
    """

    path: Optional[pathlib.Path] = None
    timeout: Optional[int] = None


@dataclasses.dataclass
class BiopaxData:  # pylint: disable=too-many-instance-attributes
    """
    BioPAX data configuration.

    Parameters:
        endpoint:
            An optional URL to a SPARQL endpoint through which to query BioPAX
            data, such as a local Fuseki server.

        include_default_owl_files:
            If True, include the default BioPAX files for the configured
            species. These files will be downloaded if necessary.

        custom_owl_files:
            A list of paths to custom OWL files in the BioPAX level-3 format to
            use, either with or without the default files depending on the value
            of include_default_owl_files.

        include_complexes:
            If True, include complexes and their components.

        include_member_entities:
            If True, include member entities (as defined by BioPAX).

        keep_unknown:
            If True, items for which a pathway or location is unknown will be
            kept when filtering by pathway and/or location.

        locations:
            Either a list of BioPax entity locations, or a path to a plaintext
            file with one location per line. See `metabograph --list-locations`
            for the complete list.

        pathways:
            Either a list of BioPAX pathways, or a path to a plaintext file with
            one pathways per line. See `metabograph --list-pathways` for the
            complete list.

        species:
            The target species. It must be one supported by BioPAX. See
            `metabograph --list-species` for the complete list.
    """

    custom_owl_files: list[pathlib.Path] = dataclasses.field(default_factory=list)
    endpoint: Optional[str] = None
    include_complexes: bool = True
    include_default_owl_files: bool = True
    include_member_entities: bool = False
    keep_unknown: bool = False
    locations: Optional[list[str] | str] = None
    pathways: Optional[list[str] | str] = None
    species: str = "homo sapiens"

    def __post_init__(self):
        self.custom_owl_files = [pathlib.Path(p) for p in self.custom_owl_files]

        for field in ("locations", "pathways"):
            value = getattr(self, field)
            if isinstance(value, str):
                LOGGER.info("Loading %s from %s", field, value)
                path = pathlib.Path(value)
                values = set(path.read_text(encoding="utf-8").splitlines())
                values = sorted(val for val in values if val)
                setattr(self, field, values)


@dataclasses.dataclass
class ConfigData:
    """
    Main configuration.

    Parameters:
        cache:
            Cache configuration.

        biopax:
            BioPAX configuration.
    """

    biopax: BiopaxData = dataclasses.field(default_factory=BiopaxData)
    cache: CacheData = dataclasses.field(default_factory=CacheData)

    def __post_init__(self):
        if isinstance(self.cache, dict):
            self.cache = CacheData(**self.cache)  # pylint: disable=not-a-mapping

        if isinstance(self.biopax, dict):
            self.biopax = BiopaxData(**self.biopax)  # pylint: disable=not-a-mapping


class Config:
    """
    Graph configuration.
    """

    def __init__(self, path: str | pathlib.Path = None, data: dict = None):
        """
        Args:
            path:
                The path to a YAML configuration file that should be loaded.

            data:
                A dict of keyword parameters for instantiating an instance of
                ConfigData.
        """
        if data is None:
            data = {}
        self.__dict__["data"] = ConfigData(**data)

        self.__dict__["path"] = None
        if path is not None:
            self.load(path)

    def load(self, path: str | pathlib.Path):
        """
        Load a configuration file.
        """
        self.__dict__["path"] = pathlib.Path(path).resolve()
        with self.path.open("rb") as handle:
            self.data = ConfigData(**yaml.safe_load(handle))

    def resolve_path(self, path: str | pathlib.Path) -> pathlib.Path:
        """
        Resolve a path relative to the configuration file's path if the path is
        set.

        Args:
            path:
                The path to resolve.

        Returns:
            The resolved path.
        """
        if self.path:
            return self.path.parent.joinpath(path)
        return pathlib.Path(path).resolve()

    def asdict(self):
        """
        Return the dict representing the current ConfigData object.
        """
        return dataclasses.asdict(self.data)

    def __getattr__(self, key: str) -> Any:
        try:
            return self.__dict__[key]
        except KeyError:
            return getattr(self.data, key)

    def __setattr__(self, key: str, value: Any):
        if key == "data":
            self.__dict__[key] = value
        else:
            setattr(self.data, key, value)

    def get_documented_yaml(self):
        """
        Return the current object as a commented YAML document.
        """
        return DataclassDocumenter(ConfigData).get_yaml()



#!/usr/bin/env python3
"""
Exceptions.
"""


class MetabographException(Exception):
    """
    Base class for custom exceptions raised by this package.
    """


class MetabographIOError(MetabographException):
    """
    Metabograph IO error.
    """


class MetabographRuntimeError(MetabographException):
    """
    Metabograph Runtime error.
    """


class MetabographConfigError(MetabographException):
    """
    Metabograph configuration error.
    """



#!/usr/bin/env python3
"""Utility functions."""


import hashlib
import logging
import os.path
import pathlib
from typing import Any

import pandas as pd


DEFAULT_HASH_ALGORITHM = "sha256"
LOGGER = logging.getLogger(__name__)


def dict_from_dataframe_columns(dataframe: pd.DataFrame, key_col: str, val_col: str):
    """
    Get a dict mapping key in one column of a dataframe to values in another
    column.

    Args:
        dataframe:
            The input dataframe.

        key_col:
            The name of the column with the keys.

        val_col:
            The name of the column with the corresponding values.

    Returns:
        A Python dict.
    """
    return dataframe[[key_col, val_col]].set_index(key_col).to_dict()[val_col]


def hash_data(data: Any, algorithm: str = DEFAULT_HASH_ALGORITHM):
    """
    Hash data.

    Args:
        data:
            The data to hash. If not a byte string then it will be converted to one.

        algorithm:
            The hashing algorithm to use.

    Returns:
        The hexdigest of the data and the algorithm used.
    """
    if not isinstance(data, bytes):
        if not isinstance(data, str):
            data = str(data)
        data = bytes(data, encoding="utf-8")
    return hashlib.new(algorithm, data).hexdigest(), algorithm


def hash_file(path: str | pathlib.Path, algorithm: str = DEFAULT_HASH_ALGORITHM):
    """
    Hash a file.

    Args:
        path:
            The path to the file.

        algorithm:
            The hashing algorithm to use.

    Returns:
        The hexdigest of the file and the algorithm used.
    """
    path = pathlib.Path(path).resolve()
    LOGGER.debug("Hashing %s with %s", path, algorithm)
    with path.open("rb") as handle:
        return hashlib.file_digest(handle, algorithm).hexdigest(), algorithm


def is_older_or_missing(dst: str | pathlib.Path, src: str | pathlib.Path):
    """
    Check if the destination path is older than the source path or missing.

    Args:
        dst:
            The destination path.

        src:
            The source path.

    Returns:
        True if the destination path is older or missing, else False
    """
    dst = pathlib.Path(dst).resolve()
    src = pathlib.Path(src).resolve()
    try:
        LOGGER.debug("Stat'ing %s", dst)
        dst_mtime = dst.stat().st_mtime
    except FileNotFoundError:
        return True
    LOGGER.debug("Stat'ing %s", src)
    return dst_mtime < src.stat().st_mtime


def get_common_prefix(items):
    """
    Get the common prefix of an iterable of strings.

    Args:
        items:
            The items to parse.

    Returns:
        The common string prefix.
    """
    items = [str(i) for i in items]
    return os.path.commonprefix(items)



#!/usr/bin/env python3
"""
Common global variables and functions.
"""

import datetime
import logging
import pathlib
import urllib.parse

import requests

from .exception import MetabographIOError

LOGGER = logging.getLogger(__name__)

# Internal name for cache directories and other uses.
NAME = "metabograph"

# HTTP header time format.
RFC_2616_FMT = "%a, %d %b %Y %H:%M:%S GMT"


def download(
    url: str,
    path: str | pathlib.Path,
    append_name: bool = False,
    timeout=10,
    force=False,
):
    """
    Download a URL to a path.

    Args:
        url:
            The URL to download.

        path:
            The output path.

        append_name:
            If True, append the URL's filename to the path.

        timeout:
            The timeout for remote requests, in seconds.

        force:
            By default, existing local files will only be overwritten if the
            remote server reports a newer modification time (or no modification
            time at all). Set this option to True to force a download without
            checking for remote modification times.

    Returns:
        The output path.
    """
    path = pathlib.Path(path).resolve()
    if append_name:
        path /= pathlib.Path(urllib.parse.urlparse(url).path).name
    if not force and path.exists():
        LOGGER.debug("Emitting HEAD request for %s", url)
        response = requests.head(url, timeout=timeout)
        try:
            remote_mtime = datetime.datetime.strptime(
                RFC_2616_FMT, response.headers["last-modified"]
            )
        except KeyError:
            pass
        else:
            local_mtime = datetime.datetime.fromtimestamp(path.stat().st_mtime)
            if remote_mtime <= local_mtime:
                return path
    path.parent.mkdir(parents=True, exist_ok=True)
    LOGGER.info("Downloading %s to %s", url, path)
    with requests.get(url, stream=True, timeout=timeout) as stream:
        try:
            stream.raise_for_status()
            with path.open("wb") as handle:
                for chunk in stream.iter_content(chunk_size=8192):
                    handle.write(chunk)
        except IOError as err:
            raise MetabographIOError(err) from err
    return path



#!/usr/bin/env python3
"""
SPARQL query functions.
"""

import hashlib
import logging

from rdflib.plugins.sparql import prepareQuery
from rdflib.plugins.sparql.algebra import translateAlgebra

LOGGER = logging.getLogger(__name__)

# Common SPARQL prefixes.
COMMON_PREFIXES = """
PREFIX bio: <http://purl.org/vocab/bio/0.1/>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
PREFIX owl: <http://www.w3.org/2002/07/owl#>
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX bp3: <http://www.biopax.org/release/biopax-level3.owl#>
PREFIX reactome: <http://identifiers.org/reactome/>
PREFIX up: <http://purl.uniprot.org/core/>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
PREFIX pubchem: <http://rdf.ncbi.nlm.nih.gov/pubchem/compound/>
PREFIX fn: <http://www.w3.org/2005/xpath-functions#>
""".lstrip()


def canonicalize(query):
    """
    Canonicalize a SPARQL query. This uses rdflib's SPARQL parsing and algebra
    translation functions. The returned query string may be longer than the
    input string so this function should not be used to shorten queryies.

    Args:
        query:
            The input SPARQL query string.

    Returns:
        The canonicalized query string.
    """
    return translateAlgebra(prepareQuery(query))


def hash_query(query):
    """
    Get a hash value for a SPARQL query. This is used to cache query results.

    Args:
        query:
            The input query.

    Returns:
        A hash string.
    """
    query = canonicalize(query).encode("utf-8")
    hsh = hashlib.sha256()
    hsh.update(query)
    return hsh.hexdigest()



#!/usr/bin/env python3
"""Run the Fuseki server."""

import argparse
import logging
import os
import pathlib
import shlex
import subprocess
import sys
import tempfile

from .biopax.query_manager import BiopaxQueryManager
from .config import Config
from .exception import MetabographException, MetabographRuntimeError
from .logging import configure_logging


LOGGER = logging.getLogger(__name__)
# The environment variable with the path to the installed Fuseki directory.
FUSEKI_HOME_VAR = "FUSEKI_HOME"


def run_fuseki_server(config):
    """
    Context manager to launch a Fuseki server
    """
    bdm = BiopaxQueryManager(config)
    fuseki_home = os.getenv(FUSEKI_HOME_VAR)
    if fuseki_home is None:
        raise MetabographRuntimeError(
            f"The {FUSEKI_HOME_VAR} environment variable is not set."
        )
    fuseki_home = pathlib.Path(fuseki_home).resolve()

    with tempfile.TemporaryDirectory() as tmp_dir:
        owl_paths = bdm.get_owl_files(tmp_dir)
        file_args = (f"--file={p}" for p in owl_paths)
        cmd = [
            str(fuseki_home / "fuseki-server"),
            *file_args,
            "/",
        ]
        LOGGER.info("Starting Fuseki server: %s", " ".join(shlex.quote(w) for w in cmd))
        LOGGER.info("Press <ctrl>+c to exit.")
        try:
            subprocess.run(cmd, check=True, cwd=tmp_dir)
        except subprocess.CalledProcessError as err:
            raise MetabographRuntimeError(err) from err


def main(args=None):
    """
    Run the Fuseki server.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "config", type=pathlib.Path, help="The Metabograph configuration file."
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Show debug messages."
    )
    pargs = parser.parse_args(args=args)
    configure_logging(level=logging.DEBUG if pargs.verbose else logging.INFO)
    config = Config(path=pargs.config)
    run_fuseki_server(config)


def run_main(args=None):
    """
    Wrapper around main with error handling.
    """
    try:
        main(args=args)
    except KeyboardInterrupt:
        pass
    except MetabographException as err:
        sys.exit(err)


if __name__ == "__main__":
    run_main()



#!/usr/bin/env python3
"""
Create NetworkX graphs from BioPAX data.
"""

import enum
import logging
from collections.abc import Iterable
from typing import Self, Any

import networkx as nx
import pandas as pd

from ..config import Config
from ..utils import dict_from_dataframe_columns
from .query_manager import BiopaxQueryManager


LOGGER = logging.getLogger(__name__)


@enum.unique
class Direction(enum.Enum):
    """
    Directions for some participant relations.
    """

    REVERSIBLE = enum.auto()
    LEFT_TO_RIGHT = enum.auto()
    RIGHT_TO_LEFT = enum.auto()

    def __str__(self) -> str:
        return self.name.replace("_", "-")

    @classmethod
    def from_str(cls, string: str) -> Self:
        """
        Convert a string to an instance of Direction.

        Args:
            string:
                The string to convert.

        Returns:
            The corresponding Direction.
        """
        if isinstance(string, cls):
            return string
        if not isinstance(string, str):
            string = str(string)
        return cls[string.upper().replace("-", "_")]


class BiopaxGraphGenerator:
    """
    Create NetworkX graphs from BioPAX data.
    """

    def __init__(self, *, config: Config = None, bqm: BiopaxQueryManager = None):
        """
        Args:
            config:
                An instance of Config. If not given, then the bqm parameter is
                required and the configuration from the BiopaxQueryManager
                instance will be used.

            bqm:
                An instance of BiopaxQueryManager. If not given, one will be
                instantiated from the given Config instance.
        """
        if bqm is None:
            bqm = BiopaxQueryManager(config)
        self.config = bqm.config
        self.bqm = bqm
        self._cached_data = {}

    def get_node_identifier(self, node: str) -> str:
        """
        Get a node identifier for the given node. This is sometimes required to
        ensure that the same entity is recognized as such when input data
        assigns different identifiers to the same entity. Subclass this class
        and override this method to handle custom identifiers.

        Args:
            node:
                The node identifier, usually the entity string.

        Returns:
            The possibly modified node identifier to use for nodes in the graph.
        """
        return node

    def get_custom_node_attributes(self, node: str) -> dict[str, Any]:
        """
        Get custom node attributes. Override in a subclass for custom user data.

        Args:
            node:
                The target node.

        Returns:
            A dict of custom node attributes to add to the given node.
        """
        return {}

    @staticmethod
    def _log_graph_addition(number: int, singular: str, plural: str = None):
        """
        Internal function for logging additions to the graph.

        Args:
            number:
                The number of elements added.

            singular:
                The display name for one element.

            plural:
                The display name for more than one element. If None, an "s" will
                be added to the singular name.
        """
        if plural is None:
            plural = f"{singular}s"
        LOGGER.info(
            "Adding %d %s to graph", number, singular if number == 1 else plural
        )

    def _add_and_log_graph_edges(
        self, graph: nx.Graph, edges: pd.Series, *args: str, **kwargs: str
    ):
        """
        Add edges to the graph and log how many where added.

        Args:
            graph:
                The nx.Graph object.

            edges:
                A Pandas series of edges to add via the nx.Graph.add_edges_from
                method.

            *args:
                Keyword arguments that will be passed through to
                _log_graph_addition after the number argument.

            **kwargs:
                Keyword arguments passed through to _log_graph_addition.
        """
        if edges.shape[0] > 0:
            self._log_graph_addition(edges.shape[0], *args, **kwargs)
            edges = edges.apply(
                lambda x: (
                    self.get_node_identifier(x[0]),
                    self.get_node_identifier(x[1]),
                    x[2],
                )
            )
            graph.add_edges_from(edges)

    @staticmethod
    def _keep_any_isin(
        data: pd.DataFrame, cols: list[str], items: Iterable[str]
    ) -> pd.DataFrame:
        """
        Keep rows for which at least one of the given columns contains a value
        in the given list of items.

        Args:
            data:
                The Dataframe to filter.

            cols:
                The list of column names to check.

            items:
                The iterable of values to check.

        Returns:
            The filtered dataframe.
        """
        if not isinstance(items, set):
            items = set(items)
        keep = None
        for col in cols:
            if keep is None:
                keep = data[col].isin(items)
            else:
                keep |= data[col].isin(items)
        return data[keep]

    @property
    def location_data(self):
        """
        The dataframe of location data.
        """
        key = "location_data"
        if self._cached_data[key] is None:
            self._cached_data[key] = self.bqm.query_locations()
        return self._cached_data[key]

    def filter_by_location(
        self, data: pd.DataFrame, ent_cols: list[str]
    ) -> pd.DataFrame:
        """
        Filter entities by their location.

        Args:
            data:
                The dataframe to filter.

            end_cols:
                The column names of the entities.

        Returns:
            The filtered dataframe.
        """
        # All entities are assocated with locations.
        locations = self.config.biopax.locations
        if locations:
            LOGGER.info("Filtering nodes by locations: %s", ", ".join(locations))
            loc_data = self.location_data
            known_loc = set(loc_data["entity"])
            selected_by_loc = set(
                loc_data[loc_data["location_name"].isin(locations)]["entity"]
            )
            keep = selected_by_loc
            if self.config.biopax.keep_unknown:
                keep |= set(data[ent_cols].values) - known_loc
            data = self._keep_any_isin(data, ent_cols, keep)
        return data

    @property
    def pathway_data(self):
        """
        The dataframe of pathway data.
        """
        key = "pathway_data"
        if self._cached_data[key] is None:
            self._cached_data[key] = self.bqm.query_pathways()
        return self._cached_data[key]

    def filter_by_pathway(
        self, data: pd.DataFrame, pathway_cols: list[str]
    ) -> pd.DataFrame:
        """
        Filter interactions by pathway.

        Args:
            data:
                The dataframe to filter.

            end_cols:
                The column names of the entities.

        Returns:
            The filtered dataframe.
        """
        # Only interactions are associated with pathways.
        pathways = self.config.biopax.pathways
        if pathways:
            LOGGER.info("Filtering interactions by pathways: %s", ", ".join(pathways))
            pathway_data = self.bqm.query_pathways()
            known_pathway_comps = set(pathway_data["component"])
            # Selected by pathway.
            keep = set(
                pathway_data[pathway_data["pathway_name"].isin(pathways)]["component"]
            )
            if self.config.biopax.keep_unknown:
                keep |= set(data[pathway_cols]) - known_pathway_comps
            data = self._keep_any_isin(data, pathway_cols, keep)

        return data

    def filter_controls(
        self, controls: pd.DataFrame, interaction_participants: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Filter controlled-controller relations to those for which the controlled
        entity is included in the interaction participants.

        Args:
            controls:
                The controls dataframe.

            interaction_participants:
                The interaction participants dataframe.
        """
        keep = controls["controlled"].isin(interaction_participants["interaction"])
        return controls[keep]

    def filter_complexes(
        self, complex_components: pd.DataFrame, interaction_participants: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Filter complex components to those containing the selected interaction
        participants. All components of each complex containing an interaction
        participant entity will be kept. This is used to indirectly filter by
        pathways.


        Args:
            complex_components:
                The complex components dataframe.

            interaction_participants:
                The interaction participants dataframe.
        """
        component_to_complex = dict_from_dataframe_columns(
            complex_components, "component", "complex"
        )
        complexes = (
            interaction_participants["entity"]
            .apply(lambda x: component_to_complex.get(x, pd.NA))
            .dropna()
        )
        complexes = pd.concat([complexes, interaction_participants["entity"]]).unique()
        keep = complex_components["complex"].isin(complexes)
        return complex_components[keep]

    def filter_member_entities(
        self, member_entities: pd.DataFrame, interaction_participants: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Filter member_entities to those containing the selected interaction
        participants. All members with a transitive member relation to one of
        the interaction participants will be keps. This is used to indirectly
        filter by pathways.

        Args:
            member_entities:
                The member entities dataframe.

            interaction_participants:
                The interaction participants dataframe.
        """
        unsorted = member_entities["a"] > member_entities["b"]
        member_entities[["a", "b"]][unsorted] = member_entities[unsorted][["b", "a"]]
        member_entities = member_entities.sort_values(by=["a", "b"])
        # Partition the member entities into sets based on the transitive
        # relation. The previous sorting guarantees that previous members will
        # be found.
        member_sets = []
        for row in member_entities.itertuples():
            ent_a = row.a
            ent_b = row.b
            for mset in member_sets:
                if ent_a in mset:
                    mset.add(ent_b)
                    break
                if ent_b in mset:
                    mset.add(ent_a)
                    break
            else:
                member_sets.append(set((ent_a, ent_b)))
        keepers = set()
        for mset in member_sets:
            if interaction_participants["entity"].isin(mset).any():
                keepers.update(mset)
        # Only one column needs to be checked because of the transitive relation.
        keep = member_entities["a"].isin(keepers)
        return member_entities[keep]

    def get_graph(self) -> nx.Graph:
        """
        Get the graph for the current configuration.
        """
        graph = nx.DiGraph()
        node_set = set()
        edge_ops = []

        # All entities have a location so they can be filter by it directly.
        # Pathways are more complicated because only interactions are associated
        # directly with them. To filter other entities by pathway, we first
        # filter all of the interaction participants by their associated
        # interaction. Other entities such as controlers, complexes and member
        # entities are then filtered by the interations and physical entities
        # remaining in interaction_participants after these filters have been
        # applied.
        interaction_participants = self.bqm.query_interaction_participants()
        interaction_participants = self.filter_by_location(
            interaction_participants, ["interaction", "entity"]
        )
        interaction_participants = self.filter_by_pathway(
            interaction_participants, ["interaction"]
        )
        node_set.update(interaction_participants["interaction"])
        node_set.update(interaction_participants["entity"])
        edge_ops.append((self._add_participant_edges, interaction_participants))

        controls = self.bqm.query_controls()
        if not controls.empty:
            controls = self.filter_by_location(controls, ["controlled", "controller"])
            # Indirect pathways filtering.
            controls = self.filter_controls(controls, interaction_participants)
            #  The tiler ensures that "controlled" is already included.
            node_set.update(controls["controller"])
            edge_ops.append((self._add_control_edges, controls))

        if self.config.biopax.include_complexes:
            complex_components = self.bqm.query_complex_components()
            if not complex_components.empty:
                complex_components = self.filter_by_location(
                    complex_components, ["complex", "component"]
                )
                # Indirect pathway filtering.
                complex_components = self.filter_complexes(
                    complex_components, interaction_participants
                )
                node_set.update(complex_components["complex"])
                node_set.update(complex_components["component"])
                edge_ops.append((self._add_complex_component_edges, complex_components))

        if self.config.biopax.include_member_entities:
            member_entities = self.bqm.query_member_entities()
            if not member_entities.empty:
                member_entities = self.filter_by_location(member_entities, ["a", "b"])
                # Indirect pathway filtering.
                member_entities = self.filter_member_entities(
                    member_entities, interaction_participants
                )
                node_set.update(member_entities["a"])
                node_set.update(member_entities["b"])
                edge_ops.append((self._add_member_entity_edges, member_entities))

        self._add_nodes(graph, node_set)
        for edge_op in edge_ops:
            edge_op[0](graph, *edge_op[1:])
        return graph

    def _add_nodes(self, graph: nx.Graph, nodes: set[str]):
        """
        Add nodes to the graph.

        Args:
            graph:
                The graph instance.

            nodes:
                The set of entities to add.
        """
        entity_to_display_names, entity_to_names = self.bqm.entity_to_names_mappers
        entity_to_location = self.bqm.entity_to_location_name_mapper
        entity_to_type = self.bqm.entity_to_type_mapper
        entity_to_simplified = self.bqm.entity_to_simplified_id
        self._log_graph_addition(len(nodes), "node")
        for node in nodes:
            dnames = entity_to_display_names.get(node)
            dnames = (
                ";".join(dnames) if dnames else entity_to_simplified.get(node, node)
            )
            names = entity_to_names.get(node)
            names = ";".join(names) if names else ""
            attrs = {
                "type": entity_to_type[node],
                "display_names": dnames,
                "names": names,
                "location": entity_to_location.get(node, "?"),
            }
            attrs.update(self.get_custom_node_attributes(node))
            graph.add_node(self.get_node_identifier(node), **attrs)

    def _add_participant_edges(
        self, graph: nx.Graph, interaction_participants: pd.DataFrame
    ):
        """
        Get the participant edge from a row of the interaction participants
        dataframe.

        Args:
            graph:
                The graph instance.

            interaction_participants:
                The dataframe of participant relations between interactions and
                entities.
        """
        edge_data = pd.DataFrame()
        edge_data["node_1"] = interaction_participants["interaction"]
        edge_data["node_2"] = interaction_participants["entity"]
        edge_data["part_name"] = interaction_participants["participant"].apply(
            lambda x: str(x).rsplit("#", 1)[1].lower()
        )
        edge_data["coefficient"] = interaction_participants["coefficient"].astype(str)

        # The participant relations "left" and "right" can be used for reactions
        # that run left-to-right, right-to-left or both (reversible). The edge
        # directions must consider all cases even if the current version of
        # Reactome only contains interactions that run left-to-right.
        left_or_right = (edge_data["part_name"] == "left") | (
            edge_data["part_name"] == "right"
        )
        # The conversion directions can be LEFT-TO-RIGHT, RIGHT-TO-LEFT, or REVERSIBLE.
        conv_dirs = self.bqm.query_conversion_directions()
        # Map each conversion to its direction.
        if conv_dirs.empty:
            conv_dict = {}
        else:
            conv_dict = pd.Series(
                conv_dirs["direction"].values, index=conv_dirs["conversion"]
            ).to_dict()
        directed = interaction_participants[left_or_right]
        edge_data.loc[left_or_right, "direction"] = directed.apply(
            (
                lambda x: Direction.from_str(
                    conv_dict.get(x["interaction"], Direction.LEFT_TO_RIGHT)
                )
            ),
            axis=1,
        )
        self._add_left_and_right_edges(graph, edge_data[left_or_right])

        # Consider other edges as bidirectional.
        others = edge_data[~left_or_right]
        self._add_and_log_graph_edges(
            graph,
            others.apply(
                (
                    lambda x: (
                        x["node_1"],
                        x["node_2"],
                        {"type": x["part_name"], "coefficient": x["coefficient"]},
                    )
                ),
                axis=1,
            ),
            "bidirectional left edge",
        )
        self._add_and_log_graph_edges(
            graph,
            others.apply(
                (
                    lambda x: (
                        x["node_2"],
                        x["node_1"],
                        {"type": x["part_name"], "coefficient": x["coefficient"]},
                    )
                ),
                axis=1,
            ),
            "bidirectional right edge",
        )

    def _add_left_and_right_edges(self, graph: nx.Graph, directed: pd.DataFrame):
        """
        Add directed edges for the left and right participant relations.

        Args:
            graph:
                The graph instance.

            directed:
                The extract of the participant dataframe with "left" or "right"
                participant relations.
        """
        rev = directed["direction"] == Direction.REVERSIBLE
        l_to_r = (directed["direction"] == Direction.LEFT_TO_RIGHT) | rev
        r_to_l = (directed["direction"] == Direction.RIGHT_TO_LEFT) | rev
        left = directed["part_name"] == "left"
        right = directed["part_name"] == "right"

        # For left-to-right and reversible interactions, add edges
        # node_1 -left-> interaction -right-> node_2.
        self._add_and_log_graph_edges(
            graph,
            directed[l_to_r & left].apply(
                (
                    lambda x: (
                        x["node_2"],
                        x["node_1"],
                        {"type": "left", "coefficient": x["coefficient"]},
                    )
                ),
                axis=1,
            ),
            "left-to-right left edge",
        )
        self._add_and_log_graph_edges(
            graph,
            directed[l_to_r & right].apply(
                (
                    lambda x: (
                        x["node_1"],
                        x["node_2"],
                        {"type": "right", "coefficient": x["coefficient"]},
                    )
                ),
                axis=1,
            ),
            "left-to-right right edge",
        )

        # For right-to-left and reversible interactions, add edges
        # node_1 <-left- interaction <-right- node_2.
        self._add_and_log_graph_edges(
            graph,
            directed[r_to_l & right].apply(
                (
                    lambda x: (
                        x["node_2"],
                        x["node_1"],
                        {"type": "right", "coefficient": x["coefficient"]},
                    )
                ),
                axis=1,
            ),
            "right-to-left right edge",
        )
        self._add_and_log_graph_edges(
            graph,
            directed[r_to_l & left].apply(
                (
                    lambda x: (
                        x["node_1"],
                        x["node_2"],
                        {"type": "left", "coefficient": x["coefficient"]},
                    )
                ),
                axis=1,
            ),
            "right-to-left left edge",
        )

    def _add_control_edges(self, graph: nx.Graph, controls: pd.DataFrame):
        """
        Add control edges.

        Args:
            graph:
                The graph instance.

            controls:
                The dataframe of control relationships.
        """
        edge_data = pd.DataFrame()
        edge_data["node_1"] = controls["controller"]
        edge_data["node_2"] = controls["controlled"]
        edge_data["type"] = controls["controlType"]
        self._add_and_log_graph_edges(
            graph,
            edge_data.apply(
                (
                    lambda x: (
                        x["node_1"],
                        x["node_2"],
                        {"type": f"control:{x['type']}"},
                    )
                ),
                axis=1,
            ),
            "control edge",
        )

    def _add_complex_component_edges(
        self, graph: nx.Graph, complex_components: pd.DataFrame
    ):
        """
        Add component edges for complexes.

        Args:
            graph:
                The graph instance.

            complex_components:
                The dataframe with complex component relations.
        """
        edge_data = pd.DataFrame()
        edge_data["node_1"] = complex_components["complex"]
        edge_data["node_2"] = complex_components["component"]
        edge_data["coefficient"] = complex_components["coefficient"].astype(str)
        self._add_and_log_graph_edges(
            graph,
            edge_data.apply(
                (
                    lambda x: (
                        x["node_1"],
                        x["node_2"],
                        {"type": "complex:component", "coefficient": x["coefficient"]},
                    )
                ),
                axis=1,
            ),
            "complex to component edge",
        )
        self._add_and_log_graph_edges(
            graph,
            edge_data.apply(
                (
                    lambda x: (
                        x["node_2"],
                        x["node_1"],
                        {"type": "component:complex", "coefficient": x["coefficient"]},
                    )
                ),
                axis=1,
            ),
            "component to complex edge",
        )

    def _add_member_entity_edges(self, graph: nx.Graph, member_entities: pd.DataFrame):
        """
        Add member entity edges.

        Args:
            graph:
                The graph instance.

            member_entities:
                The dataframe of member entity relationships.
        """
        edge_data = pd.DataFrame()
        edge_data["node_1"] = member_entities["a"]
        edge_data["node_2"] = member_entities["b"]
        self._add_and_log_graph_edges(
            graph,
            edge_data.apply(
                (lambda x: (x["node_1"], x["node_2"], {"type": "member"})), axis=1
            ),
            "member entity edge",
        )
        self._add_and_log_graph_edges(
            graph,
            edge_data.apply(
                (lambda x: (x["node_2"], x["node_1"], {"type": "member"})), axis=1
            ),
            "member edge",
        )



#!/usr/bin/env python
"""
Example of BiopaxGraphGenerator subclass for custom BioPAX data.
"""

import logging
import pathlib
import urllib.request

import networkx

from metabograph.biopax.query_manager import BiopaxQueryManager
from metabograph.biopax.graph_generator import BiopaxGraphGenerator
from metabograph.config import Config
from metabograph.logging import configure_logging


THIS_PATH = pathlib.Path(__file__).resolve()
OWL_FILE_URL = "https://www.ebi.ac.uk/biomodels/services/download/get-files/MODEL1401240000/2/MODEL1401240000-biopax3.owl"
LOGGER = logging.getLogger(__name__)


class BGGSertbas2014(BiopaxGraphGenerator):
    """
    Custom subclass to force connections between reactions.
    """

    def get_node_identifier(self, node):
        """
        The same molecule maybe be produced or consumed in the same cellular
        location by different reactions. Because the input data labels each
        reactant and product of a reaction with the reaction's identifier, the
        equality of these entities is not recognized by the graph. This method
        corrects this.

        The generated identifier strips the reaction data from the identifier to
        ensure that reactions are connected via product-reactant nodes.
        """
        location = self.bqm.entity_to_location_name_mapper.get(node)
        location = f"{self.bqm.ITEM_DELIMITER}{location}" if location else ""
        try:
            name = node.rsplit("_m_", 1)[1]
        except IndexError:
            name = node.rsplit("_n_", 1)[1]
        return f"{name}{location}"

    def get_custom_node_attributes(self, node):
        """
        Add cell_type attribute based on identifier suffix ("_N" for neurons,
        "_A" for astrocytes).
        """
        if node.endswith("_N"):
            cell_type = "neuron"
        elif node.endswith("_A"):
            cell_type = "astrocyte"
        else:
            cell_type = "?"
        return {"cell_type": cell_type}


def download_data_if_missing(path):
    """
    Download BioPAX data to local OWL file.
    """
    if not path.exists():
        LOGGER.info("Downloading %s", OWL_FILE_URL)
        path.parent.mkdir(parents=True, exist_ok=True)
        with urllib.request.urlopen(OWL_FILE_URL) as hin, path.open("wb") as hout:
            while True:
                buf = hin.read(0x1000)
                if not buf:
                    break
                hout.write(buf)


def main():
    """
    Main.
    """
    configure_logging(level=logging.INFO)
    dir_path = THIS_PATH.parent
    config_path = dir_path / "config.yaml"
    config = Config(config_path)

    owl_path = config.resolve_path(config.biopax.custom_owl_files[0])
    download_data_if_missing(owl_path)

    bqm = BiopaxQueryManager(config, debug=False)
    #  bqm.cache_man.clear()
    bgg = BGGSertbas2014(bqm=bqm)

    graph = bgg.get_graph()
    LOGGER.info("Generated graph: %s", graph)

    graph_path = owl_path.parent / "graph.gml"
    LOGGER.info("Saving graph to %s", graph_path)
    networkx.write_gml(graph, str(graph_path))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Print information about protein entities. This information is used to verify
that all retrieved proteins are correctly identified.
"""

import logging
import os
import pathlib

import pandas as pd
from tabulate import tabulate

from metabograph.biopax.query_manager import BiopaxQueryManager
from metabograph.config import Config
from metabograph.logging import configure_logging


LOGGER = logging.getLogger(__name__)


def print_dataframe(data, max_rows=None, max_columns=None, width=None):
    """
    Print an unabbreviated dataframe.
    """
    with pd.option_context(
        "display.max_rows",
        max_rows,
        "display.max_columns",
        max_columns,
        "display.width",
        width,
        "display.max_colwidth",
        None,
    ):
        print(data)


def save_dataframe(data, path, **kwargs):
    """
    Save a dataframe to a file.
    """
    path = pathlib.Path(path).resolve()
    LOGGER.info("Saving data to %s", path)
    data.to_csv(path, **kwargs)


def chdir_and_get_config():
    """
    Change directory and get the Config instance.
    """
    project_dir = pathlib.Path(__file__).parent.parent
    tmp_dir = project_dir / "tmp/count_proteins"
    tmp_dir.mkdir(parents=True, exist_ok=True)
    os.chdir(tmp_dir)
    config_path = project_dir / "config.yaml"
    return Config(config_path)


def main():
    """
    Main.
    """
    configure_logging(level=logging.DEBUG)
    config = chdir_and_get_config()

    bqm = BiopaxQueryManager(config, debug=True)
    #  bqm.cache_man.clear()
    prots = bqm.get_protein_entities()
    save_dataframe(prots, "proteins.csv", index=False)

    has_display_name = ~prots["display_names"].isna()
    has_uniprot_id = ~prots["uniprot"].isna()
    has_location = ~prots["location"].isna()
    has_xref = ~prots["xrefs"].isna()
    has_eref = ~prots["erefs"].isna()
    uniprot_in_eref_names = prots["eref_names"].str.contains("UniProt")
    has_member_uniprot = ~prots["member_uniprot"].isna()

    rows = [
        ("Number of rows in protein dataframe", prots.shape[0]),
        ("Unique protein entities", prots["entity"].nunique()),
        (
            "Unique protein entities with display names",
            prots[has_display_name]["entity"].nunique(),
        ),
        ("Unique display names", prots[has_display_name]["display_names"].nunique()),
        ("Unique UniProt IDs", prots[has_uniprot_id]["uniprot"].nunique()),
        ("Unique protein entities without cross-references", (~has_xref).sum()),
        ("Unique protein entities without reference entities", (~has_eref).sum()),
        (
            "Unique protein entities without display_name or reference entities",
            ((~has_display_name) & (~has_eref)).sum(),
        ),
        (
            "Entities with both a display name and a UniProt ID",
            prots[has_display_name & has_uniprot_id].shape[0],
        ),
        (
            "Entities with a display name and without a UniProt ID",
            prots[has_display_name & ~has_uniprot_id].shape[0],
        ),
        (
            "Entities without a display name and with a UniProt ID",
            prots[~has_display_name & has_uniprot_id].shape[0],
        ),
        (
            "Entities without a display name or a UniProt ID",
            prots[~has_display_name & ~has_uniprot_id].shape[0],
        ),
    ]
    rows.append(("Sum of last 4 rows", sum(item[1] for item in rows[-4:])))
    rows.extend(
        (
            (
                "Entities without a UniProt ID or any member UniProt IDs",
                prots[~has_uniprot_id & ~has_member_uniprot].shape[0],
            ),
            (
                'Entities without a UniProt ID but "UniProt" in entity reference names',
                prots[~has_uniprot_id & uniprot_in_eref_names].shape[0],
            ),
            (
                "Number of unique display names without a UniProt ID",
                prots[has_display_name & ~has_uniprot_id]["display_names"].nunique(),
            ),
            (
                "Number of unique UniProt IDs without a display name",
                prots[~has_display_name & has_uniprot_id]["uniprot"].nunique(),
            ),
            ("Entities with a location", has_location.sum()),
            ("Entities without a location", (~has_location).sum()),
        )
    )
    name_loc_keys = ["display_names", "location"]
    count_by_name_and_loc = (
        prots[has_display_name & has_location]
        .groupby(name_loc_keys)
        .count()
        .sort_values(name_loc_keys)
    )
    rows.extend(
        (
            (
                "Unique combinations of display name and location",
                count_by_name_and_loc.shape[0],
            ),
            (
                "Unique combinations of display name and location with single entity",
                count_by_name_and_loc[count_by_name_and_loc["entity"] == 1].shape[0],
            ),
            (
                "Unique combinations of display name and location with multiple entities",
                count_by_name_and_loc[count_by_name_and_loc["entity"] > 1].shape[0],
            ),
        )
    )
    rows.append(("Sum of last 2 rows", sum(item[1] for item in rows[-2:])))
    rows.extend(
        (
            (
                "Total number of entities grouped by display name and location",
                count_by_name_and_loc["entity"].sum(),
            ),
            (
                "Number of entities with unique combination of display name and location",
                count_by_name_and_loc[count_by_name_and_loc["entity"] == 1][
                    "entity"
                ].sum(),
            ),
            (
                "Number of entities that share a combination of display name and location",
                count_by_name_and_loc[count_by_name_and_loc["entity"] > 1][
                    "entity"
                ].sum(),
            ),
        )
    )
    rows.append(("Sum of last 2 rows", sum(item[1] for item in rows[-2:])))
    print(tabulate(rows, tablefmt="github", headers=["What", "Count"]))

    save_dataframe(
        count_by_name_and_loc["entity"],
        "display_name_and_loc_combination_counts.csv",
        index=True,
    )
    save_dataframe(
        count_by_name_and_loc[count_by_name_and_loc["entity"] > 1]["entity"],
        "non_unique_display_name_and_loc_combination_counts.csv",
        index=True,
    )
    orphans = prots[has_display_name & ~(has_uniprot_id | has_member_uniprot)][
        "display_names"
    ].sort_values()
    save_dataframe(orphans, "display_names_without_uniprot_id.csv", index=False)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Configure logging.
"""

import logging

LOGGER = logging.getLogger(__name__)


def configure_logging(level=logging.INFO):
    """
    Configure logging.

    Args:
        level:
            Logging level.
    """
    logging.basicConfig(
        style="{",
        format="[{asctime}] {levelname} {message}",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=level,
    )



#!/usr/bin/env python3
"""
UniProt ID mapping.
"""

import contextlib
import logging

from uniprot_id_mapping import IdMapper
from uniprot_id_mapping.exception import UniprotIdMappingException
from uniprot_id_mapping.cache import CacheManager as UIM_CacheManager

from ..cache import CacheManager
from ..exception import MetabographException


LOGGER = logging.getLogger(__name__)


class MetabographIdMapperError(MetabographException):
    """
    Custom exception raised by IdMapper.
    """


@contextlib.contextmanager
def id_mapper_context(cache_man: CacheManager):
    """
    Context manager to obtain an instance of uniprot_id_mapping.IdMapper
    configured to use Metabograph's cache directory. The context will convert
    UniprotIdMappingExceptions to MetabographIdMapperErrors

    Args:
        cache_man:
            A CacheManager instance.

    Returns:
        A configured instance of uniprot_id_mapping.IdMapper
    """
    try:
        uim_cache_man = UIM_CacheManager(
            cache_dir=cache_man.get_path("uniprot", directory=True)
        )
        id_mapper = IdMapper(uim_cache_man)
        yield id_mapper
    except UniprotIdMappingException as err:
        raise MetabographIdMapperError(err) from err



#!/usr/bin/env python3
"""
Run SPARQL queries on OWL files.
"""

import logging
import pathlib

import pandas as pd

try:
    import owlready2

    WorldType = owlready2.World
except ImportError:
    owlready2 = None
    WorldType = None  # pylint: disable=invalid-name

try:
    import rdflib
except ImportError:
    rdflib = None


from .exception import MetabographRuntimeError


LOGGER = logging.getLogger(__name__)


class OwlLoader:
    """
    Load OWL files and run SPARQL queries on them.
    """

    def __init__(self, cache_man, name, paths):
        """
        Args:
            cache_man:
                A CacheManager instance.

            name:
                A name for this loader. It will be used as a database name by
                the owlready backend to speed up queries after the initial
                loading.

            paths:
                Paths to OWL files to load.
        """
        self.cache_man = cache_man
        self.name = name
        self.paths = [pathlib.Path(p).resolve() for p in paths]
        self._owlready2_world = None
        self._rdflib_graph = None

    @property
    def owlready2_world(self) -> WorldType:
        """
        owlready2 ontology.
        """
        if not owlready2:
            return None
        if self._owlready2_world is None:
            db_path = self.cache_man.get_path(f"owlready/{self.name}.sqlite3")
            exists = db_path.exists()
            self._owlready2_world = owlready2.World(filename=db_path)
            if exists:
                LOGGER.info("Reloading cached data from %s", db_path)
            else:
                for path in self.paths:
                    LOGGER.info("Loading %s with owlready2", path)
                    self._owlready2_world.get_ontology(path.as_uri()).load()
                LOGGER.info("Caching data to %s", db_path)
                self._owlready2_world.save()
        return self._owlready2_world

    @property
    def rdflib_graph(self) -> rdflib.Graph:
        """
        Run a query using rdflib.
        """
        if not rdflib:
            return None
        if self._rdflib_graph is None:
            graph = rdflib.Graph()
            for path in self.paths:
                LOGGER.info("Loading %s with rdflib", path)
                graph.parse(str(path))
            self._rdflib_graph = graph
        return self._rdflib_graph

    def load(self):
        """
        Load the configured OWL files. Normally the OWL files are loaded lazily
        on demand but this function can be used to force loading of files from a
        temporary context.
        """
        if self.owlready2_world is None:
            if self.rdflib_graph is None:
                LOGGER.debug("Failed to load OWL files.")

    def query(self, query: str) -> pd.DataFrame:
        """
        Run a SPARQL query on the loaded ontology.

        Args:
            query:
                The SPARQL query to run.

        Returns:
            A Pandas dataframe with the results.
        """
        if not self.paths:
            return None

        graph = None
        world = self.owlready2_world
        if world:
            graph = world.as_rdflib_graph()
            backend = "owlready2"
        else:
            graph = self.rdflib_graph
            backend = "rdflib"
        if graph:
            LOGGER.debug("Running SPARQL query through %s", backend)
            rows = graph.query(query)
            LOGGER.debug("Converting query results to dataframe")
            return pd.DataFrame((row.asdict() for row in rows))

        raise MetabographRuntimeError(
            "One of the following packages is required: owlready2, rdflib"
        )



#!/usr/bin/env python3
"""
Download and query BioPAX data.

References:
* https://www.biopax.org/owldoc/Level3/
"""

import functools
import logging
import pathlib
import sys
import tempfile
import textwrap
import urllib.error
import zipfile
from typing import Any

import numpy as np
import pandas as pd
from SPARQLWrapper import SPARQLWrapper2

from ..cache import CacheManager
from ..common import download
from ..config import Config
from ..exception import MetabographIOError
from ..owl import OwlLoader
from ..sparql import COMMON_PREFIXES, hash_query
from ..utils import dict_from_dataframe_columns, is_older_or_missing, get_common_prefix
from .protein import get_protein_entities


LOGGER = logging.getLogger(__name__)


def _format_species(species: str) -> str:
    """
    Format a genus and species string for display.

    Args:
        species:
            The genus and species string, e.g. "homo sapiens".

    Returns:
        The formatted genus and species string, e.g. "Homo sapiens".
    """
    genus, species = species.replace("_", " ").split(None, 2)
    return f"{genus.title()} {species.lower()}"


def owl_file(species: str) -> str:
    """
    Get the name of the OWL file for the given species.
    """
    return f'{_format_species(species).replace(" ", "_")}.owl'


class BiopaxQueryManager:
    """
    BioPAX data manager.
    """

    ITEM_DELIMITER = ";;;"
    FIELD_DELIMITER = ":::"

    def __init__(self, config: Config, debug: bool = False):
        self.config = config
        self.debug = debug
        self.cache_man = CacheManager(self.config)
        self._owlloader = None
        self._sparqlwrapper = None
        self._constraints = None

    @property
    def biopax_dir(self) -> pathlib.Path:
        """
        The cache directory path for BioPAX data.
        """
        return self.cache_man.get_path("biopax", directory=True)

    @property
    def species(self) -> str:
        """
        The configured species.
        """
        return self.config.biopax.species

    @property
    def endpoint(self) -> str:
        """
        The configured SPARQL endpoint. It may be None.
        """
        return self.config.biopax.endpoint

    @property
    def level_3_owl_file(self) -> pathlib.Path:
        """
        Get the path to the BioPAX level 3 OWL file. This may be a
        user-configured path. It not set, the biopax.org data will be used. The
        file will be downloaded if necessary.
        """
        path = self.biopax_dir / "biopax-level3.owl"
        if not path.exists():
            url = "https://www.biopax.org/release/biopax-level3.owl"
            download(url, path)
        return path

    @property
    def reactome_archive_path(self) -> pathlib.Path:
        """
        The path to the zipped Reactome BioPAX archive with the specied data.
        The will will be downloaded if missing.
        """
        zip_path = self.biopax_dir / "biopax.zip"
        if not zip_path.exists():
            url = "https://reactome.org/download/current/biopax.zip"
            download(url, zip_path)
        return zip_path

    def extract_species_owl_file(self, dir_path: str | pathlib.Path) -> pathlib.Path:
        """
        Extract a species OWL file to a directory. The data will be downloaded
        if necessary.

        Args:
            dir_path:
                An output directory path.

        Returns:
            The path to the extract file.
        """
        dir_path = pathlib.Path(dir_path).resolve()
        dir_path.mkdir(parents=True, exist_ok=True)
        zip_path = self.reactome_archive_path
        species_file = owl_file(self.species)
        output_path = dir_path / species_file
        if is_older_or_missing(output_path, zip_path):
            LOGGER.info("Extracting %s from %s to %s", species_file, zip_path, dir_path)
            with zipfile.ZipFile(zip_path) as handle:
                handle.extract(species_file, path=dir_path)
        return output_path

    def get_owl_files(self, dir_path: str | pathlib.Path) -> list[pathlib.Path]:
        """
        Get the configured OWL files. Retrievable files may be downloaded or
        extracted to the given path. Existing files will simply return their
        paths, which may lie outside of the given directory.

        Args:
            dir_path:
                The path to a directory.

        Returns:
            The generator of OWL file paths as pathlib.Path objects.
        """
        # This file is necessary.
        yield self.level_3_owl_file
        bpconf = self.config.biopax
        if bpconf.include_default_owl_files:
            dir_path = pathlib.Path(dir_path).resolve()
            dir_path.mkdir(parents=True, exist_ok=True)
            yield self.extract_species_owl_file(dir_path)
        yield from bpconf.custom_owl_files

    def list_species(self) -> list[str]:
        """
        List the available species.
        """
        with zipfile.ZipFile(self.reactome_archive_path) as handle:
            for name in sorted(handle.namelist()):
                name = name.split(".", 1)[0]
                yield _format_species(name)

    def _query_sparqlwrapper(self, query: str) -> pd.DataFrame:
        """
        Query a SPARQL endpoint via SPARQLWrapper2.
        """
        endpoint = self.endpoint
        if self._sparqlwrapper is None:
            self._sparqlwrapper = SPARQLWrapper2(endpoint)
        self._sparqlwrapper.setQuery(query)
        try:
            LOGGER.debug("Sending SPARQL query to %s", endpoint)
            rows = self._sparqlwrapper.query().bindings
            LOGGER.debug("Converting query results to dataframe")
            return pd.DataFrame(
                [{key: value.value for (key, value) in row.items()} for row in rows]
            )
        except AttributeError:
            LOGGER.error("Query to %s failed: %s", endpoint, query)
            return None
        except urllib.error.URLError as err:
            LOGGER.error("Failed to query %s: %s", endpoint, err)
            raise MetabographIOError(err) from err

    def _query_owlloader(self, query: str) -> pd.DataFrame:
        """
        Query local OWL files.
        """
        if self._owlloader is None:
            with tempfile.TemporaryDirectory() as tmp_dir:
                paths = list(self.get_owl_files(tmp_dir))
                species_filename = pathlib.Path(owl_file(self.species)).stem.lower()
                name = f"biopax-{species_filename}"
                self._owlloader = OwlLoader(self.cache_man, name, paths)
                self._owlloader.load()
        return self._owlloader.query(query)

    @staticmethod
    def _custom_debug_msg(title: str, content: Any):
        """
        Print a custom debug message to STDERR.
        """
        width = 6
        content = str(content)
        for line in content.splitlines():
            width = max(width, len(line.rstrip()))
        line_chr = "-"
        # Ensure that even long titles are bookended by at least 3 line characters.
        header = f"{line_chr * 3}{title}{line_chr * 3}"
        header = str.center(title, width, line_chr)
        footer = line_chr * len(header)
        sys.stdout.write(f"{header}\n{content}\n{footer}\n")

    def _get_cached_query_path(self, query):
        """
        Get the path to a cached query depending on the query itself and the
        current configuration.

        Args:
            query:
                A SPARQL query string.
        """
        return (
            self.biopax_dir
            / "cached_queries"
            / self.species.replace(" ", "_")
            / f"{hash_query(query)}.csv"
        )

    def query(self, query: str) -> pd.DataFrame:
        """
        Run a SPARQL query on the loaded graph.

        Args:
            query:
                A SPARQL query string.
        """
        # Ensure that NaN and empty strings are preserved in the cached results.
        null_value = "NULL"
        query = textwrap.dedent(query).strip()
        query = f"{COMMON_PREFIXES}\n{query}"
        if self.debug:
            self._custom_debug_msg("SPARQL Query", query)
        path = self._get_cached_query_path(query)
        if path.exists():
            LOGGER.debug("Reloading cached query results from %s", path)
            try:
                data = pd.read_csv(path, na_values=[null_value], keep_default_na=False)
            except pd.errors.EmptyDataError:
                LOGGER.debug("Cached query data is empty: %s", path)
                return pd.DataFrame()
        else:
            if self.endpoint:
                data = self._query_sparqlwrapper(query)
            else:
                data = self._query_owlloader(query)
            # Convert empty fields to NaN.
            with pd.option_context("future.no_silent_downcasting", True):
                data.replace(r"^\s*$", np.nan, regex=True, inplace=True)
            LOGGER.debug("Caching query results to %s", path)
            path.parent.mkdir(parents=True, exist_ok=True)
            data.to_csv(path, index=False, na_rep=null_value)
        if self.debug:
            self._custom_debug_msg("SPARQL Query Results", data)
        return data

    def query_locations(self) -> pd.DataFrame:
        """
        Query all locations. Each entity is associated with at most one
        location.
        """
        query = """\
        SELECT DISTINCT ?entity ?location ?location_name
        WHERE {
            ?entity bp3:cellularLocation ?location.
            ?location bp3:term ?location_name.
        }
        """
        return self.query(query)

    def list_locations(self) -> list[str]:
        """
        List all known locations by name.
        """
        return sorted(self.query_locations()["location_name"].unique())

    def query_pathways(self) -> pd.DataFrame:
        """
        Query all pathways and their components.
        """
        query = """\
        SELECT DISTINCT ?pathway ?pathway_name ?component
        WHERE {
            ?pathway rdf:type bp3:Pathway;
                bp3:displayName ?pathway_name;
                bp3:pathwayComponent+ ?component.
        }
        """
        return self.query(query)

    def list_pathways(self) -> list[str]:
        """
        List all known pathways by name.
        """
        return sorted(set(self.query_pathways()["pathway_name"]))

    def query_entities(self) -> pd.DataFrame:
        """
        Query all entities and their types.
        """
        query = """\
        SELECT DISTINCT
            ?entity
            ?entity_type
        WHERE {
            ?entity rdf:type/rdfs:subClassOf* bp3:Entity;
                rdf:type ?entity_type.
        }
        """
        return self.query(query)

    def query_physical_entities(self) -> pd.DataFrame:
        """
        Query all physical entities and their references.
        """
        query = f"""\
        SELECT DISTINCT
            ?entity
            ?entity_type
            (GROUP_CONCAT(DISTINCT ?entity_name; separator="{self.ITEM_DELIMITER}") AS ?names)
            (
                GROUP_CONCAT(DISTINCT ?display_name; separator="{self.ITEM_DELIMITER}")
                AS ?display_names
            )
            (GROUP_CONCAT(DISTINCT ?eref; separator="{self.ITEM_DELIMITER}") AS ?erefs)
            (GROUP_CONCAT(DISTINCT ?eref_name; separator="{self.ITEM_DELIMITER}") AS ?eref_names)
            (GROUP_CONCAT(DISTINCT ?xref_str; separator="{self.ITEM_DELIMITER}") AS ?xrefs)
            (GROUP_CONCAT(DISTINCT ?member; separator="{self.ITEM_DELIMITER}") AS ?members)
        WHERE {{
            ?entity rdf:type/rdfs:subClassOf* bp3:PhysicalEntity;
                rdf:type ?entity_type.

            OPTIONAL {{
                ?entity bp3:name ?opt_entity_name.
            }}
            BIND(IF(BOUND(?opt_entity_name), ?opt_entity_name, "") AS ?entity_name)

            OPTIONAL {{
                ?entity bp3:displayName ?opt_display_name.
            }}
            BIND(IF(BOUND(?opt_display_name), ?opt_display_name, "") AS ?display_name)

            OPTIONAL {{
                ?entity bp3:entityReference+ ?opt_eref.
                ?opt_eref bp3:xref ?opt_xref;
                    bp3:name ?opt_eref_name.
                ?opt_xref bp3:id ?xref_id;
                    bp3:db ?xref_db.
                BIND(CONCAT(?xref_db,"{self.FIELD_DELIMITER}",?xref_id) AS ?opt_xref_str).
            }}
            BIND(IF(BOUND(?opt_eref), ?opt_eref, "") AS ?eref)
            BIND(IF(BOUND(?opt_eref_name), ?opt_eref_name, "") AS ?eref_name)
            BIND(IF(BOUND(?opt_xref_str), ?opt_xref_str, "") AS ?xref_str)

            OPTIONAL {{
                ?entity bp3:memberPhysicalEntity ?opt_member.
            }}
            BIND(IF(BOUND(?opt_member), ?opt_member, "") AS ?member)
        }}
        GROUP BY ?entity ?entity_type
        """
        return self.query(query)

    def query_interactions(self) -> pd.DataFrame:
        """
        Query all interactions.
        """
        query = """\
        SELECT DISTINCT ?interaction ?interaction_name
        WHERE {
            ?interaction rdf:type/rdfs:subClassOf* bp3:Interaction.

            OPTIONAL {
                ?interaction bp3:displayName ?opt_interaction_name.
            }
            BIND(IF(BOUND(?opt_interaction_name), ?opt_interaction_name, "") AS ?interaction_name)
        }
        """
        return self.query(query)

    def query_member_entities(self) -> pd.DataFrame:
        """
        Query all member entities.
        """
        query = """\
        SELECT DISTINCT ?a ?b
        WHERE {
            ?a bp3:memberPhysicalEntity ?b.
        }
        """
        return self.query(query)

    def query_conversion_directions(self) -> pd.DataFrame:
        """
        Query all conversion directions.
        """
        query = """\
        SELECT DISTINCT ?conversion ?direction
        WHERE {
            ?conversion rdf:type/rdfs:subClassOf* bp3:Conversion;
                bp3:conversionDirection ?direction.
        }
        """
        return self.query(query)

    def query_participants(self) -> pd.DataFrame:
        """
        Query all participant types.
        """
        query = """\
        SELECT DISTINCT ?participant
        WHERE {
            ?participant rdf:type/rdfs:subPropertyOf* bp3:participant.
        }
        """
        return self.query(query)

    def query_interaction_participants(self) -> pd.DataFrame:
        """
        Query all interaction-participant-entity relations.
        """
        query = """\
        SELECT DISTINCT ?interaction ?participant ?entity ?coefficient
        WHERE {
            ?interaction rdf:type/rdfs:subClassOf* bp3:Interaction.
            ?participant rdfs:subPropertyOf* bp3:participant.
            ?interaction ?participant ?entity.

            OPTIONAL {
                ?interaction bp3:participantStoichiometry ?opt_stoichiometry.
                ?opt_stoichiometry bp3:physicalEntity ?entity;
                    bp3:stoichiometricCoefficient ?opt_coeff.
            }
            BIND(IF(BOUND(?opt_coeff), ?opt_coeff, "") AS ?coefficient)
        }
        """
        return self.query(query)

    def query_controls(self) -> pd.DataFrame:
        """
        Query all controlled-controller relations.
        """
        query = """\
        SELECT DISTINCT ?controlled ?controller ?controlType
        WHERE {
            ?control rdf:type/rdfs:subClassOf* bp3:Control;
                bp3:controlled ?controlled;
                bp3:controller ?controller;
                bp3:controlType ?controlType.
        }
        """
        return self.query(query)

    def query_complex_components(self) -> pd.DataFrame:
        """
        Query all complex components.
        """
        query = """\
        SELECT DISTINCT ?complex ?component ?coefficient
        WHERE {
            ?complex rdf:type bp3:Complex.
            ?complex bp3:component ?component.

            OPTIONAL {
                ?complex bp3:componentStoichiometry ?opt_stoichiometry.
                ?opt_stoichiometry bp3:physicalEntity ?component;
                    bp3:stoichiometricCoefficient ?opt_coeff.
            }
            BIND(IF(BOUND(?opt_coeff), ?opt_coeff, "") AS ?coefficient)
        }
        """
        return self.query(query)

    def _map_key_to_values(
        self, data: pd.DataFrame, key_col: str, val_col: str
    ) -> pd.DataFrame:
        """
        Get a dict mapping keys to sets of values.

        Args:
            data:
                The dataframe from which to create the dict. It will be modified
                in-situ.

            key_col:
                The name of the key column.

            val_col:
                the name of the value column. The elements in this column are
                assumed to be separated by self.ITEM_DELIMITER.

        Returns:
            A dict mapping keys to sets of values.
        """
        mod_val_col = f"split_{val_col}"
        data[mod_val_col] = data[val_col].apply(
            lambda x: x.split(self.ITEM_DELIMITER) if isinstance(x, str) else pd.NA
        )
        return dict_from_dataframe_columns(data, key_col, mod_val_col)

    @functools.cached_property
    def entity_to_names_mappers(self) -> (dict[str, str], dict[str, str]):
        """
        Dicts mapping physical entities to names. The first maps entities to
        display names while the second maps entities to names. The names may be
        inconsistent due to the underlying data.

        TODO:
            Add support for other entity types.
        """
        physical_entities = self.query_physical_entities()
        data = pd.DataFrame(physical_entities[["entity", "display_names", "names"]])
        dnames = self._map_key_to_values(data, "entity", "display_names")
        names = self._map_key_to_values(data, "entity", "names")
        return dnames, names

    @functools.cached_property
    def entity_to_location_name_mapper(self) -> dict[str, str]:
        """
        A dict mapping entities to their locations.
        """
        locations = self.query_locations()
        return dict_from_dataframe_columns(locations, "entity", "location_name")

    @functools.cached_property
    def entity_to_type_mapper(self) -> dict[str, str]:
        """
        A dict mapping entities to their types.
        """
        entities = self.query_entities()
        key_col = "entity"
        val_col = "entity_type"
        data = pd.DataFrame(entities[[key_col, val_col]])
        data[val_col] = data[val_col].apply(lambda x: x.rsplit("#", 1)[1])
        return dict_from_dataframe_columns(data, key_col, val_col)

    @functools.cached_property
    def entity_to_simplified_id(self) -> dict[str, str]:
        """
        A dict mapping entities to simplified ID strings. This works by removing
        the common prefix from all entities. If there is no common prefix, the
        returned dict will be empty.
        """
        entities = self.query_entities()
        key_col = "entity"
        val_col = "simplified"
        data = pd.DataFrame(entities[key_col])
        prefix_len = len(get_common_prefix(data[key_col]))
        if prefix_len == 0:
            return {}
        data[val_col] = data[key_col].apply(lambda x: x[prefix_len:])
        return dict_from_dataframe_columns(data, key_col, val_col)


# Add methods define in other modules.
BiopaxQueryManager.get_protein_entities = get_protein_entities



#!/usr/bin/env python3
"""Metabograph command-line application."""

import argparse
import logging
import pathlib
import sys
from collections.abc import Iterable
from typing import Any

import networkx

from .biopax.graph_generator import BiopaxGraphGenerator
from .biopax.query_manager import BiopaxQueryManager
from .config import Config
from .exception import MetabographException
from .logging import configure_logging

LOGGER = logging.getLogger(__name__)


def _print_list(title: str, items: Iterable[Any]):
    """
    Print a list.

    Args:
        title:
            The title of the list.

        items:
            The list items.
    """
    print(title)
    for item in items:
        print(f"  {item}")


def main(args=None):
    """
    Main function.
    """
    parser = argparse.ArgumentParser(
        description="Generate graphs from Reactome BioPAX data."
    )
    parser.add_argument(
        "config", nargs="?", help="Path to the YAML configuration file."
    )
    parser.add_argument(
        "graph", nargs="?", help="Output path for the generated graph in GML format."
    )
    parser.add_argument(
        "--clear-cache",
        action="store_true",
        help="Clear the cache to force a refresh of query data.",
    )
    parser.add_argument(
        "--create-config",
        type=pathlib.Path,
        metavar="PATH",
        help="""
            Create a YAML configuration file at the given path. If the path is
            "-", the generated YAML will be printed to STDOUT.
        """,
    )
    parser.add_argument(
        "--list-species", action="store_true", help="List available species."
    )
    parser.add_argument(
        "--list-locations",
        action="store_true",
        help="List recognized cellular locations.",
    )
    parser.add_argument(
        "--list-pathways", action="store_true", help="List recognized pathways."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increasing logging level to DEBUG. Pass twice to also show SPARQL queries.",
    )
    pargs = parser.parse_args(args=args)
    configure_logging(level=logging.DEBUG if pargs.verbose else logging.INFO)
    config = Config(path=pargs.config)
    bqm = BiopaxQueryManager(config, debug=pargs.verbose > 1)

    if pargs.clear_cache:
        bqm.cache_man.clear()

    if pargs.list_species:
        _print_list("Species", bqm.list_species())
        return

    if pargs.list_locations:
        _print_list("Cellular Locations", bqm.list_locations())
        return

    if pargs.list_pathways:
        _print_list("Pathways", bqm.list_pathways())
        return

    if pargs.create_config:
        config_text = config.get_documented_yaml()
        if pargs.create_config.name == "-":
            print(config_text)
            return
        path = pargs.create_config.resolve()
        LOGGER.info("Creating configuration file at %s", path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(config_text, encoding="utf-8")
        return

    LOGGER.info("Generating graph.")
    bgg = BiopaxGraphGenerator(bqm=bqm)
    graph = bgg.get_graph()

    if pargs.graph:
        path = pathlib.Path(pargs.graph).resolve()
        LOGGER.info("Saving graph to %s", path)
        path.parent.mkdir(parents=True, exist_ok=True)
        networkx.write_gml(graph, str(path))
        test = networkx.read_gml(str(path))
        print("Reloaded", test)
        return

    print("Generated graph:", graph)


def run_main(args=None):
    """
    Wrapper around main with error handling.
    """
    try:
        main(args=args)
    except KeyboardInterrupt:
        pass
    except MetabographException as err:
        sys.exit(err)


if __name__ == "__main__":
    run_main()
